import tiktoken
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, AsyncGenerator
from hwhkit.utils import logger


class ModelNotFoundError(Exception):
    def __init__(self, model_name: str):
        super().__init__(f"Model '{model_name}' not found in the model instances.")


class LanguageModelProperty:
    def __init__(
            self, name: str, short_name: str, company: str, max_input_token: int, max_output_token: int,
            top_p: Optional[float] = None, top_k: Optional[int] = None, temperature: Optional[float] = None,
            timeout: int = 120, input_token_fee_pm: float = 0.0, output_token_fee_pm: float = 0.0,
            train_token_fee_pm: float = 0.0, prompt_version: int = 1, keys: List[str] = None
    ):
        self.name = name
        self.short_name = short_name
        self.company = company
        self.max_input_token = max_input_token
        self.max_output_token = max_output_token
        self.top_p = top_p
        self.top_k = top_k
        self.temperature = temperature
        self.timeout = timeout
        self.input_token_fee_pm = input_token_fee_pm
        self.output_token_fee_pm = output_token_fee_pm
        self.train_token_fee_pm = train_token_fee_pm
        self.prompt_version = prompt_version


class ModelStrategy(ABC):

    @abstractmethod
    def base_url(self) -> str: pass

    @abstractmethod
    def get_name(self) -> str: pass

    @abstractmethod
    async def gen_messages(
            self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> List[Dict[str, str]]:
        pass

    @abstractmethod
    async def chat(self, messages: List[Dict[str, str]]) -> str:
        pass

    @abstractmethod
    async def chat_stream(self, messages: List[Dict[str, str]]) -> AsyncGenerator[str, None]:
        pass

    def calculate_token_count(self, text: str, encoding_name: str = "cl100k_base") -> int:
        encoding = tiktoken.get_encoding(encoding_name)
        return len(encoding.encode(text))

    def calculate_messages_token_count(self, messages: List[Dict[str, str]], encoding_name: str = "cl100k_base") -> int:
        total_tokens = 0
        for message in messages:
            content = f"role:{message['role']} content:{message['content']}"
            total_tokens += self.calculate_token_count(content, encoding_name)
        return total_tokens

    async def resized_messages(
            self, messages: List[Dict[str, str]], encoding_name: str = "cl100k_base"
    ) -> List[Dict[str, str]]:
        if not messages:
            return []

        max_token_count = self._property.max_input_token
        total_tokens = self.calculate_messages_token_count(messages, encoding_name)
        if total_tokens <= max_token_count:
            return messages

        while total_tokens > max_token_count and len(messages) > 1:
            removed_message = messages.pop(1)
            total_tokens = self.calculate_messages_token_count(messages, encoding_name)
            logger.info(f"Removed message due to token limit: {removed_message}")

        if total_tokens > max_token_count and len(messages) > 0:
            removed_message = messages.pop(0)
            total_tokens = self.calculate_messages_token_count(messages, encoding_name)
            logger.info(f"Removed system message due to token limit: {removed_message}")
        return messages


class LanguageModel:

    def __init__(self, strategy: ModelStrategy):
        self._strategy = strategy

    async def chat(
            self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> str:
        messages = await self._strategy.gen_messages(question, system_prompt, history_messages)
        messages = await self._strategy.resized_messages(messages)
        return await self._strategy.chat(messages)

    async def chat_stream(
            self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> AsyncGenerator[str, None]:
        messages = await self._strategy.gen_messages(question, system_prompt, history_messages)
        messages = await self._strategy.resized_messages(messages)
        async for chunk in self._strategy.chat_stream(messages):
            yield chunk

    async def gen_messages(
            self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> List[Dict[str, str]]:
        return await self._strategy.gen_messages(question, system_prompt, history_messages)

    async def chat_with_message(self,  messages: List[Dict[str, str]]) -> str:
        return await self._strategy.chat(messages)

    async def chat_stream_with_message(self, messages: List[Dict[str, str]]) -> AsyncGenerator[str, None]:
        async for chunk in self._strategy.chat_stream(messages):
            yield chunk


class LanguageModelFactory:
    _model_instances: Dict[str, LanguageModel] = {}

    @classmethod
    def register_model_instance(cls, name: str, model_instance: LanguageModel):
        cls._model_instances[name] = model_instance

    @classmethod
    def get_model_instance(cls, name: str) -> LanguageModel:
        model = cls._model_instances.get(name)
        if not model:
            logger.error(f"Model '{name}' not found in the model instances.")
            raise ModelNotFoundError(name)
        return model

    @classmethod
    def list_models(cls) -> List[str]:
        return list(cls._model_instances)

    @classmethod
    def all_models(cls) -> Dict[str, LanguageModel]:
        return cls._model_instances



# !/usr/bin/env python
# -*-coding:utf-8 -*-

"""
# Time       ：2024/7/29 14:42
# Author     ：Maxwell
# Description：
"""


class HttpResult(object):

    def __init__(self, succeed: bool, content: bytes, text: str = '', message: str = ''):
        self.succeed = succeed
        self.content = content
        self.text = text
        self.message = message





import asyncio
from dataclasses import dataclass
from functools import wraps
from typing import Optional, Dict, List, Callable, Any
from aiomqtt import Client, MqttError
from hwhkit.utils import logger
from hwhkit.connection.mqtt.plugins.encryption import EncryptionPlugin


@dataclass
class MQTTConfig:
    client_id: str
    broker: str
    port: int = 1883
    username: Optional[str] = None
    password: Optional[str] = None


class MQTTAsyncClientManager:
    def __init__(self, mqtt_config="mqtt_key_pairs.yaml"):
        self.clients: Dict[str, Client] = {}
        self.topic_callbacks: Dict[str, List[Callable[[Client, str, bytes], Any]]] = {}
        self.plugins = [EncryptionPlugin(mqtt_config)]
        self.reversed_plugins = list(reversed(self.plugins))
        self.client_tasks: Dict[str, asyncio.Task] = {}

    async def add_client(self, config: MQTTConfig):
        client = Client(
            identifier=config.client_id,
            hostname=config.broker,
            port=config.port,
            username=config.username,
            password=config.password
        )
        self.clients[config.client_id] = client
        self.client_tasks[config.client_id] = asyncio.create_task(self._run_client(client, config.client_id))
        return self

    async def _run_client(self, client: Client, client_id: str):
        reconnect_interval = 5  # seconds
        while True:
            try:
                async with client:
                    logger.info(f"Client {client_id} connected successfully.")
                    await self._subscribe_to_topics(client, client_id)
                    await self._listen_messages(client, client_id)
            except MqttError as e:
                logger.error(f"Error with client {client_id}: {e}")
                logger.info(f"Reconnecting in {reconnect_interval} seconds...")
                await asyncio.sleep(reconnect_interval)
            except asyncio.CancelledError:
                logger.info(f"Client {client_id} task cancelled.")
                break

    async def _subscribe_to_topics(self, client: Client, client_id: str):
        for topic in self.topic_callbacks.keys():
            try:
                await client.subscribe(topic)
                logger.info(f"Client {client_id} subscribed to topic: {topic}")
            except MqttError as e:
                logger.error(f"Failed to subscribe to topic {topic} for client {client_id}: {e}")

    async def _listen_messages(self, client: Client, client_id: str):
        async for message in client.messages:
            logger.info(f"Received【{message.topic}】 size: {len(message.payload)}")
            await self._process_message(client, str(message.topic), message.payload)

    async def _process_message(self, client: Client, topic: str, payload: bytes):
        if topic in self.topic_callbacks:
            processed_message = payload
            for plugin in self.reversed_plugins:
                processed_message = plugin.on_message_received(topic, processed_message)
            for callback in self.topic_callbacks[topic]:
                try:
                    await callback(client, topic, processed_message)
                except Exception as e:
                    logger.error(f"Error in callback for topic {topic}: {e}")

    def topic_handler(self, topic: str):
        def decorator(func: Callable[[Client, str, bytes], Any]):
            @wraps(func)
            async def wrapper(client, topic, message):
                return await func(client, topic, message)

            if topic not in self.topic_callbacks:
                self.topic_callbacks[topic] = []
            self.topic_callbacks[topic].append(wrapper)
            logger.info(f"Added handler for topic: {topic}")
            return wrapper

        return decorator

    async def publish(self, client_id: str, topic: str, message: str, qos: int = 0, retain: bool = False):
        client = self.clients.get(client_id)
        if client:
            processed_message = message
            for plugin in self.plugins:
                processed_message = plugin.on_message_published(topic, processed_message)

            try:
                await client.publish(topic, processed_message.encode(), qos=qos, retain=retain)
                logger.info(f"Client {client_id} published message to {topic}")
            except MqttError as e:
                logger.error(f"Client {client_id} failed to publish message to {topic}: {e}")
        else:
            raise ValueError(f"Client with ID {client_id} not found.")

    async def run(self):
        try:
            await asyncio.gather(*self.client_tasks.values())
        except asyncio.CancelledError:
            logger.info("MQTT client manager is shutting down...")
        finally:
            await self.disconnect_all()

    async def disconnect_all(self):
        for client_id, task in self.client_tasks.items():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
            logger.info(f"Disconnected client {client_id}")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.disconnect_all()




import os
import base64
import time

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import padding
from cryptography.exceptions import InvalidKey, UnsupportedAlgorithm


class AESCipher:
    BLOCK_SIZE = algorithms.AES.block_size
    KEY_SIZE = 32
    IV_SIZE = int(BLOCK_SIZE/8)

    def __init__(self, key: bytes = None):
        if key:
            if len(key) != self.KEY_SIZE:
                raise ValueError(f"Key must be {self.KEY_SIZE} bytes long.")
            self.key = key
        else:
            self.key = os.urandom(self.KEY_SIZE)

    def encrypt(self, plaintext: str) -> str:
        try:
            iv = os.urandom(self.IV_SIZE)
            padder = padding.PKCS7(self.BLOCK_SIZE).padder()
            padded_data = padder.update(plaintext.encode('utf-8')) + padder.finalize()
            cipher = Cipher(algorithms.AES(self.key), modes.CBC(iv))
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(padded_data) + encryptor.finalize()
            return base64.b64encode(iv + ciphertext).decode('utf-8')
        except ValueError as e:
            raise ValueError(f"Encryption failed: {str(e)}")
        except InvalidKey:
            raise ValueError("Encryption failed: Invalid key.")
        except UnsupportedAlgorithm:
            raise ValueError("Encryption failed: Unsupported algorithm.")
        except Exception as e:
            raise ValueError(f"Encryption failed: {str(e)}")

    def decrypt(self, ciphertext: str) -> str:
        try:
            encrypted_data = base64.b64decode(ciphertext.encode('utf-8'))
            iv = encrypted_data[:self.IV_SIZE]
            ciphertext = encrypted_data[self.IV_SIZE:]
            cipher = Cipher(algorithms.AES(self.key), modes.CBC(iv))
            decryptor = cipher.decryptor()
            padded_data = decryptor.update(ciphertext) + decryptor.finalize()
            unpadder = padding.PKCS7(self.BLOCK_SIZE).unpadder()
            plaintext = unpadder.update(padded_data) + unpadder.finalize()
            return plaintext.decode('utf-8')
        except ValueError as e:
            raise ValueError(f"Decryption failed: {str(e)}")
        except InvalidKey:
            raise ValueError("Decryption failed: Invalid key.")
        except UnsupportedAlgorithm:
            raise ValueError("Decryption failed: Unsupported algorithm.")
        except Exception as e:
            raise ValueError(f"Decryption failed: {str(e)}")

    def get_key(self) -> str:
        return base64.b64encode(self.key).decode('utf-8')

    @classmethod
    def get_key_bytes(cls, base64_bytes: bytes) -> bytes:
        base64_str = base64_bytes.decode("utf-8")
        return base64.b64decode(base64_str)

if __name__ == '__main__':

    times = 100000
    s = time.time_ns()
    for i in range(0, times):
        aes_cipher = AESCipher()
        plaintext = "Hello, AES encryption!"
        ciphertext = aes_cipher.encrypt(plaintext)
        # print("Ciphertext:", ciphertext)
        decrypted_text = aes_cipher.decrypt(ciphertext)
        # print("Decrypted text:", decrypted_text)
        # print("AES Key:", aes_cipher.get_key())
    e = time.time_ns()
    print(f"use time: {(e-s)/1000_000}")




# !/usr/bin/env python
# -*-coding:utf-8 -*-

"""
# Time       ：2025/1/3 19:28
# Author     ：Maxwell
# Description：
"""
import random
from typing import Dict, List, Optional, AsyncGenerator
from hwhkit.llm.models import LanguageModel


class MOERouter:
    def __init__(self, model_instances: Dict[str, LanguageModel]):
        self.model_instances = model_instances

    async def chat(
        self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> str:
        selected_model = await self.select_model(question)
        return await selected_model.chat(question, system_prompt, history_messages)

    async def chat_stream(
        self, question: str, system_prompt: Optional[str] = None, history_messages: Optional[List[Dict]] = None
    ) -> AsyncGenerator[str, None]:
        selected_model = self.select_model(question)
        async for chunk in selected_model.chat_stream(question, system_prompt, history_messages):
            yield chunk

    async def select_model(self, question: str) -> LanguageModel:
        return random.choice(list(self.model_instances.values()))



import aiofiles

class AsyncFileHandler:
    def __init__(self, file_path):
        self.file_path = file_path

    async def read_file(self, binary=False):
        mode = 'rb' if binary else 'r'
        async with aiofiles.open(self.file_path, mode=mode) as file:
            content = await file.read()
            return content

    async def write_file(self, content, binary=False):
        mode = 'wb' if binary else 'w'
        async with aiofiles.open(self.file_path, mode=mode) as file:
            await file.write(content)

    async def append_file(self, content, binary=False):
        mode = 'ab' if binary else 'a'
        async with aiofiles.open(self.file_path, mode=mode) as file:
            await file.write(content)
